local copImg = actor.LoadSprite("image/cop.png", { 128, 128 })
local kingImg = actor.LoadSprite("image/king.png", { 192, 192 })
local beamImg = actor.LoadSprite("image/beam.png", { 128, 256 })
local candyImg = actor.LoadSprite("image/candy.png", { 64, 32 })
local lollyImg = actor.LoadSprite("image/lolly.png", { 32, 64 })
local judgeImg = actor.LoadSprite("image/judge.png", { 128, 128 })
local circleImg = actor.LoadSprite("image/circle.png", { 128, 128 })
local twinkleImg = actor.LoadSprite("image/twinkle.png", { 64, 64 })

local function Lolly(pos, vel)
	local obj = {
		z = 100,
		pos = pos,
		velocity = vel,
		class = "Lolly",
		sprite = lollyImg,
		bb = actor.Square(4),
		offset = { x = -8, y = -24 },
		spriteSize = { x = 16, y = 32 },
	}
	obj = actor.Create(obj)
	local angle = vector.Angle(vel) - 90
	util.RotateTile(obj.tile, angle + 180)

	local rAngle = vector.Radians(angle - 180)
	eapi.AnimateAngle(obj.tile, eapi.ANIM_LOOP,
			  vector.null, rAngle, 0.5, -math.random())
end

local srcPos = { x = -12, y = -12 }
local srcSize = { x = 24, y = 24 }
local dstPos = { x = -1, y = -1 }
local dstSize = { x = 2, y = 2 }
local srcColor = { r = 0.65, g = 0.3, b = 1.0, a = 1.0 }
local dstColor = { r = 0.65, g = 0.3, b = 1.0, a = 0.0 }
local function LollyDust(pos, vel, ttl, angle)
	local body = eapi.NewBody(gameWorld, pos)
	local tile = eapi.NewTile(body, srcPos, srcSize, twinkleImg, 100)
	eapi.SetColor(tile, srcColor)
	eapi.AnimateSize(tile, eapi.ANIM_CLAMP, dstSize, ttl, 0)
	eapi.AnimatePos(tile, eapi.ANIM_CLAMP, dstPos, ttl, 0)
	eapi.AnimateColor(tile, eapi.ANIM_CLAMP, dstColor, ttl, 0)
	util.AnimateRotation(tile, 1.0 + math.random(), angle)	
	util.DelayedDestroy(body, ttl)
	eapi.SetVel(body, vel)
	return body
end

local dustVel = { x = 0, y = 100 }
local function LollyBurst(enemy)
	local pos = eapi.GetPos(enemy.body)
	local acc = vector.Normalize(eapi.GetVel(enemy.body), 800)

	for i = 1, 10, 1 do
		local angle = 360 * math.random()
		local vel = vector.Rotate(dustVel, angle)
		vel = vector.Scale(vel, 0.5 * (1 + math.random()))
		local body = LollyDust(pos, vel, 0.5, angle)
		eapi.SetAcc(body, acc)
	end
end

local function Reset()
	player.hits = player.hits + 1
	timing.Reset()
end

local function HitPlayer(pShape, eShape)
	local enemy = actor.store[eShape]
	eapi.PlaySound(gameWorld, "sound/lick.ogg")
	LollyBurst(enemy)
	actor.Delete(enemy)
	Reset()
end

actor.SimpleCollide("Player", "Lolly", HitPlayer)

local function Eyes(parent, offset, radius, distance, depth, target)
	local body = eapi.NewBody(parent, offset)
	local size = { x = radius, y = radius }
	local pos = { x = -distance, y = -0.5 * radius }
	local tile = eapi.NewTile(body, pos, size, circleImg, depth)
	local eye = { tile = tile }
	eapi.SetColor(tile, util.Gray(0))
	local function Follow()
		local bodyPos = eapi.GetPos(body, gameWorld)
		local targetPos = eapi.GetPos(target, gameWorld)
		eye.angle = vector.Angle(vector.Sub(bodyPos, targetPos))
		util.RotateTile(tile, eye.angle)
		eye.timer = eapi.AddTimer(body, 0.02, Follow)
	end
	Follow()
	return eye
end

local function BeamsCircle(obj, phase, delay)
	local z = -100
	local ttl = 0.5
	local body = obj.body
	local srcPos = { x = -1, y = 0 }
	local srcSize = { x = 2, y = 4 }
	local dstPos = { x = -64, y = 512 }
	local dstSize = { x = 128, y = 512 }

	local color = { r = 1.0, g = 1.0, b = 0.8, a = 0.01 }

	for angle = 0, 359, 10 do
		local tile = eapi.NewTile(body, srcPos, srcSize, beamImg, z)
		eapi.AnimateSize(tile, eapi.ANIM_LOOP, dstSize, ttl, delay)
		eapi.AnimatePos(tile, eapi.ANIM_LOOP, dstPos, ttl, delay)
		util.AnimateRotation(tile, 32, angle + phase)
		eapi.SetColor(tile, color)
	end
end

local function Beams(obj)
	local phase = 0
	for delay = 0.0, 0.99, 0.2 do
		BeamsCircle(obj, phase, -delay)
		phase = 5 - phase
	end
end

local function Stop(obj)
	eapi.PlayMusic("sound/ussr.ogg", nil, 0.9, 1.0)
	util.Delay(staticBody, 1.0, timing.Advance, "no")
	eapi.SetVel(obj.body, vector.null)
	eapi.SetAcc(obj.body, vector.null)
end

local function King()
	local obj = { z = 200,
		      class = "Blocker",
		      sprite = kingImg,
		      vel = vector.null,
		      pos = { x = 0, y = 350 },
		      offset = { x = -96, y = -96 } }
	obj = actor.Create(obj)
	actor.MakeShape(obj, { l = -32, r = 24, b = -70, t = 60 })
	actor.MakeShape(obj, { l = 24, r = 70, b = -50, t = 40 })
	actor.MakeShape(obj, { l = 60, r = 100, b = 0, t = 60 })
	actor.MakeShape(obj, { l = -80, r = -32, b = -55, t = 32 })
	actor.MakeShape(obj, { l = -100, r = -70, b = -15, t = 50 })
	npc.obj[0] = obj
	npc.king = obj
	Beams(obj)
end

local function StopHelper(obj)
	eapi.SetVel(obj.body, vector.null)
	if not npc.helpersArrived then
		npc.helpersArrived = true
		timing.Advance("no")
	end
end

local function Helper(img, dir)
	local obj = { z = 180,
		      sprite = img,
		      class = "Blocker",
		      vel = vector.null,
		      bb = actor.Square(40),
		      pos = { x = dir * 500, y = 150 },
		      offset = { x = -64, y = -64 } }
	obj = actor.Create(obj)
	npc.obj[dir] = obj
	eapi.SetVel(obj.body, { x = -dir * 250, y = 0 })
	util.Delay(obj.body, 1.0, StopHelper, obj)
	eapi.AnimateAngle(obj.tile, eapi.ANIM_CLAMP,
			  vector.null, 2 * dir * math.pi, 1, 0)
end

local function GoHelpers(body)
	Helper(copImg, -1)
	Helper(judgeImg, 1)
	timing.Pause()
end

local function KingEye(pos)
	return Eyes(npc.king.body, pos, 24, 20, npc.king.z - 1, player.me.body)
end

local function GoKing()
	eapi.SetVel(npc.king.body, { x = 0, y = -400 })
	eapi.SetAcc(npc.king.body, { x = 0, y = 400 })
	util.Delay(npc.king.body, 1.0, Stop, npc.king)

	local z = npc.king.z - 2
	local size = { x = 80, y = 32 }
	local offset = { x = -45, y = -24 }
	local tile = eapi.NewTile(npc.king.body, offset, size, util.white, z)
	npc.king.white = tile

	npc.king.eye = { }
	npc.king.eye[1] = KingEye({ x = 18, y = -6 })
	npc.king.eye[-1] = KingEye({ x = -26, y = -8 })
end

local function Orbit(obj)
	actor.Link(obj, npc.obj[npc.KING])
	eapi.SetStepC(obj.body, eapi.STEPFUNC_ROT, 2)
	util.AnimateRotation(obj.tile, 1)
end

local function StopAll()
	util.Map(function(x) eapi.SetVel(x.body, vector.null) end, npc.obj)
	Orbit(npc.obj[npc.JUDGE])
	Orbit(npc.obj[npc.COP])
	timing.Advance("no")
end

local function GoAll()
	eapi.SetVel(npc.obj[npc.COP].body, { x = 150, y = -50 })
	eapi.SetVel(npc.obj[npc.KING].body, { x = 0, y = -50 })
	eapi.SetVel(npc.obj[npc.JUDGE].body, { x = -150, y = -50 })
	eapi.AddTimer(staticBody, 1.0, StopAll)
	timing.Pause()
end

local function GetAngle()
	local copBody = npc.obj[npc.COP].body
	local kingBody = npc.obj[npc.KING].body
	return vector.Angle(eapi.GetPos(copBody, kingBody))
end

local function WaitTime()
	local angle = GetAngle()
	angle = (angle < 0) and angle or (180 - angle)
	return math.abs(vector.Radians(angle / 2))
end

local function GoAway(obj)
	local dir = util.Sign(eapi.GetPos(obj.body, gameWorld).x)
	eapi.SetStepC(obj.body, eapi.STEPFUNC_STD)
	eapi.SetVel(obj.body, { x = 250 * dir, y = 0 })
	eapi.SetAcc(obj.body, vector.null)
	actor.DelayedDelete(obj, 2)
end

local function Blow(tile, dstSize, dstPos)
	eapi.AnimatePos(tile, eapi.ANIM_CLAMP, dstPos, 2, 0)
	eapi.AnimateSize(tile, eapi.ANIM_CLAMP, dstSize, 2, 0)
end

local srcColor = { r = 0.2, g = 1.0, b = 0, a = 1.0 }
local dstColor = { r = 0.2, g = 1.0, b = 0, a = 0.0 }
local dustGravity = { x = 0, y = -200 }
local function StopKingDust(body, tile)
	eapi.AnimateColor(tile, eapi.ANIM_CLAMP, dstColor, 1, 0)
	eapi.SetAcc(body, dustGravity)
end

local srcPos = { x = -8, y = -8 }
local srcSize = { x = 16, y = 16 }
local function KingDust(pos, vel)
	local body = eapi.NewBody(gameWorld, pos)
	local tile = eapi.NewTile(body, srcPos, srcSize, twinkleImg, 200)
	util.AnimateRotation(tile, 1.0 + math.random(), angle)
	util.Delay(body, 1.0, StopKingDust, body, tile)
	util.DelayedDestroy(body, 2)
	eapi.SetColor(tile, srcColor)
	eapi.SetAcc(body, vector.Add(dustGravity, vector.Scale(vel, -1)))
	eapi.SetVel(body, vel)
	return body
end

local function PopulateKingDust(king)
	local base = { x = 0, y = 1 }
	local basePos = eapi.GetPos(king.body)
	eapi.PlaySound(gameWorld, "sound/clap.ogg")
	util.Delay(staticBody, 2.0, timing.Advance, "no")
	for i = 1, 500, 1 do
		local angle = 360 * math.random()
		local distance = math.sqrt(math.random())
		local dir = vector.Scale(vector.Rotate(base, angle), distance)
		local pos = vector.Add(basePos, vector.Scale(dir, 64))
		local vel = vector.Scale(dir, 1000)
		KingDust(pos, vel)
	end
end

local function BlowUpKing()
	local king = npc.obj[npc.KING]
	actor.DeleteShape(king)
	eapi.Destroy(king.white)
	eapi.Destroy(king.tile)
	PopulateKingDust(king)
	for i = -1, 1, 2 do
		eapi.Destroy(king.eye[i].tile)
	end
end

local function FixupCandyAngle(angle)
	if math.abs(angle) > 90 then
		return -util.Sign(angle) * (180 - math.abs(angle))
	else
		return angle
	end
end

local function CandyText(pos, vel)
	local obj = {
		z = 190,
		pos = pos,
		velocity = vel,
		sprite = candyImg,
		offset = { x = -32, y = -16 },
	}
	obj = actor.Create(obj)
	actor.DelayedDelete(obj, 2)
	util.RotateTile(obj.tile, FixupCandyAngle(vector.Angle(vel)))
end

local function EmitCandyStream()
	local counter = 150
	local direction = 1
	local function CandyStream()
		local obj = npc.king
		local pos = eapi.GetPos(obj.body)
		local vel = { x = direction * 400, y = 0 }
		local angle = 10 * (0.5 - math.random())
		CandyText(vector.Rnd(pos, 16), vector.Rotate(vel, angle))
		direction = -direction
		if counter > 0 then
			eapi.AddTimer(obj.body, 0.01, CandyStream)
			counter = counter - 1
		end
	end
	CandyStream()
end

local function BeGone()
	EmitCandyStream()
	GoAway(npc.obj[npc.COP])
	GoAway(npc.obj[npc.JUDGE])
	local king = npc.obj[npc.KING]
	Blow(king.tile, { x = 256, y = 256 }, { x = -128, y = -128 })
	Blow(king.white, { x = 110, y = 48 }, { x = -64, y = -36 })
	eapi.AddTimer(king.body, 2, BlowUpKing)
	eapi.FadeMusic(2.0)
end

local function RollEye(i, eye)
	util.AnimateRotation(eye.tile, 0.5 * i, eye.angle)
	eapi.CancelTimer(eye.timer)
end

local function EndIt()
	for i = -1, 1, 2 do RollEye(i, npc.king.eye[i]) end
	eapi.AddTimer(staticBody, WaitTime(), BeGone)
	timing.Pause()
end

local function ObjPos(index)
	return eapi.GetPos(npc.obj[index].body, gameWorld)
end

npc = {
	COP = -1,
	KING = 0,
	JUDGE = 1,
	obj = { },
	Eyes = Eyes,
	King = King,
	EndIt = EndIt,
	GoAll = GoAll,
	Lolly = Lolly,
	GoKing = GoKing,
	ObjPos = ObjPos,
	GoHelpers = GoHelpers,
	beamImg = beamImg,
}
