local lollySpeed = 250
local lollyVel = { x = 0, y = -lollySpeed }

local function Aim(index)
	local dir = vector.Sub(player.Pos(), npc.ObjPos(index or npc.KING))
	return vector.Normalize(dir, lollySpeed)
end

local function isNumber(arg)
	return (type(arg) == "number")
end

local function LollyVel(arg)
	return vector.Rotate(lollyVel, arg)
end

local function Lolly(index, arg)
	npc.Lolly(npc.ObjPos(index), isNumber(arg) and LollyVel(arg) or arg)
end

local function KingLolly(arg)
	Lolly(npc.KING, arg)
end

local function EmitRay(i, offset, body)
	local vel = Aim(i)
	local displace = 0
	local function Emit()
		displace = (displace + 1) % 3
		local angle = 5 * (offset - 1) + displace - 1
		Lolly(i, vector.Rotate(vel, angle))
	end
	util.Repeater(Emit, 5, 0.1, body, 0)
end

local function SimpleRays(set)
	local offset = 0
	local function Rays(body)
		for _, i in pairs(set) do EmitRay(i, offset, body) end
		util.Delay(body, 0.3, Rays, body)
		offset = (offset + 1) % 3
	end
	return Rays
end

local function EmitFan(i, step, body)
	local vel = Aim(i)
	local function Emit()
		for angle = -2 * step, 2 * step, step do
			Lolly(i, vector.Rotate(vel, angle))
		end
	end
	util.Repeater(Emit, 7, 0.1, body, 0)
end

local function SimpleFans(set)
	local step = 20
	local function Fans(body)
		for _, i in pairs(set) do EmitFan(i, step, body) end
		util.Delay(body, 0.9, Fans, body)
		step = 50 - step
	end
	return Fans
end

local function EmitCloud(dir, index)
	for i = 1, 30, 1 do
		local jitter = 0.8 + 0.4 * math.random()
		local angle = 30 * (0.5 - math.random())
		local vel = vector.Scale(dir, jitter)
		Lolly(index, vector.Rotate(vel, angle))
	end
end

local function CouldAtAngle(angle)
	return function(index)
		EmitCloud(vector.Rotate(Aim(index), angle), index)
	end
end

local function SimpleCloud(set)
	local function Cloud(body)
		util.Map(CouldAtAngle(0), set)
		util.Delay(body, 0.8, Cloud, body)
	end
	return Cloud
end

local function MoreClouds(set, delay)
	local function Cloud(body)
		for angle = -60, 60, 60 do
			util.Map(CouldAtAngle(angle), set)
		end
		util.Delay(body, delay, Cloud, body)
	end
	return Cloud
end

local function SimpleArcs()
	local phase = 0
	local mu = 0.0025
	local function Arcs(body)
		for alpha = 0, 359, 30 do
			for beta = 1, 20, 2 do
				local angle = alpha + beta + phase
				local vel = vector.Rotate(lollyVel, angle)
				KingLolly(vector.Scale(vel, 1.0 + mu * beta))
			end
		end
		util.Delay(body, 0.7, Arcs, body)
		phase = 15 - phase
		mu = -mu
	end
	return Arcs
end

local HELPER_ARC_SIZE = 10
local function OneArc(index, phase)
	for alpha = 0, 359, 3 * HELPER_ARC_SIZE do
		for beta = 0, HELPER_ARC_SIZE, 2 do
			Lolly(index, index * (alpha + beta + phase))
		end
	end
end

local function HelperArcs()
	local index = 1
	local phase = 0
	local function Arcs(body)
		OneArc(index, phase)
		index = -index
		phase = (phase + 2) % 360
		util.Delay(body, 0.5, Arcs, body)
	end
	return Arcs
end

local function Phyllotaxis()
	local vel = lollyVel
	local function Emit(body)
		vel = vector.Rotate(vel, util.fibonacci)
		util.Delay(body, 0.02, Emit, body)
		KingLolly(vel)
	end
	return Emit
end

local scale = 0.02
local function GetNoise(x, y, z)
        return eapi.Fractal(scale * x, scale * y, scale * z, 1, 2)
end

local function PerlinNoise()
	local step = 2
	local omega = 0
        local delta = 0
        local progress = 0
        local arm = { x = 128, y = 0 }
        local function Perlin(body)
                for angle = 0, 359, step do
			local alpha = angle + delta + step * omega
                        local pos = vector.Rotate(arm, alpha)
                        local size = GetNoise(pos.x, pos.y, progress)
                        if size > 0.5 then KingLolly(alpha) end
                end
		omega = (omega + util.golden) % 1.0
                delta = (delta + step) % 360
                progress = progress + 12

		eapi.AddTimer(body, 0.05, Perlin)
        end
	return Perlin
end

GetNoise(0, 0, 0) -- to init table

local function Diamonds()
	local counter = 0
	local flip = true
	local shape = { 0, 4, 7, 9, 11, 9, 7, 4, 0,
			-1, -1, -1, -1, -1, -1, -1 }
	local function Emit(body)
		local flip = true
		for delta = 0, 359, 20 do
			flip = not flip
			local index = flip and counter or (counter + 8)
			local gamma = shape[(index % #shape) + 1]
			if gamma >= 0 and (flip or counter > 1) then
				KingLolly(delta - gamma)
				KingLolly(delta + gamma)
			end
		end
		counter = counter + 1
		util.Delay(body, 0.1, Emit, body)
	end
	return Emit
end

local function ShootThree(index)
	local away = vector.Sub(npc.ObjPos(index), npc.ObjPos(npc.KING))
	away = vector.Normalize(away, lollySpeed)
	for angle = 0, 359, 90 do
		Lolly(index, vector.Rotate(away, angle))
	end
end

local function ShootThreeWay()
	local counter = 0
	local function Emit(body)
		ShootThree(npc.COP)
		ShootThree(npc.JUDGE)
		counter = (counter + 1) % 5
		local delay = (counter == 0) and 0.2 or 0.02
		util.Delay(body, delay, Emit, body)
	end
	return Emit
end

local function CarouselLine(index)
	local away = vector.Sub(npc.ObjPos(index), npc.ObjPos(npc.KING))
	local vel = vector.Rotate(away, -90)
	for q = 0.8, 1.2, 0.1 do
		Lolly(index, vector.Normalize(vel, q * lollySpeed))
	end
end

local function Carousel()
	local counter = 0
	local function Emit(body)
		CarouselLine(npc.COP)
		CarouselLine(npc.JUDGE)
		counter = (counter + 1) % 10
		if counter == 0 then
			EmitCloud(Aim(npc.KING), npc.KING)
		end
		util.Delay(body, 0.1, Emit, body)
	end
	return Emit
end

local function Burst(index)
	local aim = Aim(index)
	for q = 0.9, 1.1, 0.05 do
		local vel = vector.Normalize(aim, lollySpeed * q)
		for angle = 0, 359, 30 do
			Lolly(index, vector.Rotate(vel, angle))
		end
	end
end

local function Fireworks()
	local flip = false
	local function Emit(body)
		Burst(flip and npc.COP or npc.JUDGE)
		util.Delay(body, 0.2, Emit, body)
		flip = not flip
	end
	return Emit
end

local king = { npc.KING }

local all = { npc.COP, npc.KING, npc.JUDGE }

timing.Init({ SimpleRays(king),
	      SimpleCloud(king),
	      SimpleFans(king),
	      MoreClouds(king, 0.9),
	      SimpleArcs(),
	      PerlinNoise(),
	      Phyllotaxis(),
	      Diamonds(),
	      npc.GoHelpers,
	      SimpleRays(all),
	      SimpleCloud(all),
	      SimpleFans(all),
	      MoreClouds(all, 1.5),
	      HelperArcs(),
	      npc.GoAll,
	      ShootThreeWay(),
	      Fireworks(),
	      Carousel(),
	      npc.EndIt,
	    })
