local LEFT   = 1
local RIGHT  = 2
local UP     = 3
local DOWN   = 4

actor.Cage("Blocker", 0.6, 0.5)

actor.SimpleCollide("Blocker", "Player", actor.Blocker)

local playerImg = actor.LoadSprite("image/player.png", { 64, 64 })

local function GetDirection(moves)
        return { x = moves[RIGHT] - moves[LEFT], y = moves[UP] - moves[DOWN] }
end

local function SetSpeed(obj)
        eapi.SetVel(obj.body, vector.Normalize(obj.vel, obj.speed))
end

local function Move(obj, axis)
        return function(keyDown)
                obj.moves[axis] = (keyDown and 1) or 0
                obj.vel = GetDirection(obj.moves)
                SetSpeed(obj)
        end
end

local function Enable(obj, Fn)
        input.Bind("Up", true, Fn(obj, UP))
        input.Bind("Down", true, Fn(obj, DOWN))
        input.Bind("Left", true, Fn(obj, LEFT))
        input.Bind("Right", true, Fn(obj, RIGHT))
end

local function First(obj, axis)
	local Fn = Move(obj, axis)
	return function(keyDown)
		player.RemoveText()
		Enable(obj, Move)
		Fn(keyDown)
	end
end

local haloUp = { r = 1.0, g = 0.2, b = 0.2, a = 0.4 }
local haloDown = { r = 1.0, g = 0.2, b = 0.2, a = 0.0 }

local haloPos = { x = -2, y = 4 }
local haloSize = { x = 1, y = 2 }
local function Halo(obj)
	local index = 1
	local z = obj.z - 1
	local img = npc.beamImg

	for angle = 0, 359, 10 do
		local tile = eapi.NewTile(obj.body, haloPos, haloSize, img, z)
		util.RotateTile(tile, angle + 185)
		eapi.SetColor(tile, haloDown)
		player.halo[index] = tile
		index = index + 1
	end
end

local function ResetHaloTile(tile)
	eapi.SetPos(tile, haloPos)
	eapi.SetSize(tile, haloSize)
	eapi.SetColor(tile, haloDown)
end

local pos = { x = -6, y = 16 }
local size = { x = 12, y = 32 }
local function AnimateHalo(index, t, delay)
	local tile = player.halo[index]
	ResetHaloTile(tile)
	eapi.AnimateColor(tile, eapi.ANIM_CLAMP, haloUp, t, delay)
	eapi.AnimateSize(tile, eapi.ANIM_CLAMP, size, t, delay)
	eapi.AnimatePos(tile, eapi.ANIM_CLAMP, pos, t, delay)
end

local function Reset()
	local stretch = 4
	local count = #player.halo / 2
	local t = player.timeout / (count + stretch - 1)

	for i = 1, count, 1 do
		local num = (i - 1)
		local j = 2 * count - num
		local delay = num * t
		AnimateHalo(i, t * stretch, delay)
		AnimateHalo(j, t * stretch, delay)
	end
end

local function Create()
	local obj = { z = 0,
		      speed = 300,
		      class = "Player",
		      sprite = playerImg,
		      vel = vector.null,
		      moves = { 0, 0, 0, 0 },
		      pos = { x = 0, y = -128 },
		      offset = { x = -32, y = -32 },
		      bb = { l = -5, r = 6, b = -12, t = -4 } }
	obj = actor.Create(obj)
	Enable(obj, First)
	player.me = obj
	Halo(obj)

	local size = { x = 32, y = 11 }
	local offset = { x = -16, y = -6 }
	local tile = eapi.NewTile(obj.body, offset, size, util.white, -2)
	eapi.SetColor(tile, { r = 1.0, g = 0.4, b = 0.4 })

	npc.Eyes(obj.body, { x = 8, y = 2 }, 10, 7, -1, npc.king.body)
	npc.Eyes(obj.body, { x = -10, y = 0 }, 10, 7, -1, npc.king.body)
end

local function Pos()
	return eapi.GetPos(player.me.body, gameWorld)
end

local function Pause()
	util.Map(ResetHaloTile, player.halo)
end

player = {
	hits = 0,
	Pos = Pos,
	halo = { },
	Reset = Reset,
	Pause = Pause,
	Create = Create,
	timeout = 7,
}
